/****************************************************************************
 *
 * PAM module to create user credentials from keytab file
 *
 * Lots of code is derived from Heimdal's kinit and
 * Balazs Gal <balsa@rit.bme.hu> pam_krb5 PAM module
 *
 * Author: Andreas Haupt
 *
 ****************************************************************************/

#define SERVICENAME "pam_krb5keytab"

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD

#ifndef KRB5_SUCCESS
#define KRB5_SUCCESS 0
#endif

#define MAXLEN 512

#include <pwd.h>
#include <security/pam_modules.h>
#include <krb5.h>

#ifdef KRB4
#include <krb.h>
#include <krb-protos.h>
#include <kafs.h>
#endif

/* generate kerberos tickets and AFS token from keytab */
int initialize_creds(pam_handle_t *, int, const char **);
/* delete credentials cache files */
int delete_creds(pam_handle_t *);
/* build ccache name */
int get_cc_name(char *, const char *, ...);
/* get uid and gid from user */
int get_pw(const char *, uid_t *, gid_t *);
/* log to syslog */
void log(int priority, const char *, ...);

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh,
	int flags, int argc, const char **argv)
{
	log(LOG_DEBUG, "called pam_sm_authenticate %d", flags);
/*	return pam_sm_setcred(pamh, flags || PAM_ESTABLISH_CRED, argc, argv); */
	return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamh,
	int flags, int argc, const char **argv)
{
	log(LOG_DEBUG, "called pam_sm_acct_mgmt %d", flags);
	return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh,
	int flags, int argc, const char **argv)
{
	log(LOG_DEBUG, "called pam_sm_setcred %d", flags);

	if ( flags & PAM_ESTABLISH_CRED )
		return initialize_creds(pamh, argc, argv);
	if ( flags & PAM_DELETE_CRED )
		return delete_creds(pamh);
	return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh,
	int flags, int argc, const char **argv)
{
	log(LOG_DEBUG, "called pam_sm_chauthtok %d", flags);
	return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh,
	int flags, int argc, const char **argv)
{
	log(LOG_DEBUG, "called pam_sm_open_session %d", flags);
	return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh,
	int flags, int argc, const char **argv)
{
	log(LOG_DEBUG, "called pam_sm_close_session %d", flags);
	return delete_creds(pamh);
}

/* end PAM section */

int initialize_creds(pam_handle_t *pamh, int argc, const char **argv)
{
	int krc = KRB5_SUCCESS, prc = PAM_SUCCESS, i, uid, gid;
	char *user = NULL, *ktfile = NULL;
	char krb5cc[MAXLEN], cache_file5[MAXLEN];
#ifdef KRB4
	char cache_file4[MAXLEN], krb4cc[MAXLEN];
	krb5_boolean v4_convert = 0;
#endif
	krb5_context context;
	krb5_principal principal;
	krb5_ccache ccache;
	krb5_keytab kt;
	krb5_creds cred;
#ifdef KRB4
	CREDENTIALS v4cred;
#endif
	krb5_addresses addresses;
	krb5_deltat start_time = 0;
	krb5_get_init_creds_opt opt;

	log(LOG_DEBUG, "called initialize_creds");

	/* fetching config parameter */
	for ( i = 0; i < argc; i++ )
	{
		if (! strncmp("keytab=", argv[i], 7)) ktfile = argv[i] + 7;
	}

	if ( (krc = krb5_init_context(&context)) != KRB5_SUCCESS )
		return PAM_SYSTEM_ERR;
	krb5_init_ets(context);

	if ( (prc = pam_get_item(pamh, PAM_USER, &user)) != PAM_SUCCESS )
	{
		log(LOG_CRIT, "could not determine user name");
		prc = PAM_USER_UNKNOWN;
		goto out;
	}
	log(LOG_DEBUG, "getting ticket for user %s", user);

	/* important to do this: segfaults otherwise */
	memset(&addresses, 0, sizeof(addresses));
	memset(&cred, 0, sizeof(cred));
	if ( (krc = krb5_parse_name(context, user, &principal)) != KRB5_SUCCESS )
	{
		log(LOG_CRIT, "%s building user principal %s",
			 error_message(krc), user);
		prc = PAM_SYSTEM_ERR;
		goto out;
	}
	if ( !get_pw(user, &uid, &gid) )
	{
		log(LOG_CRIT, "error getting uid/gid for user");
		prc = PAM_USER_UNKNOWN;
		goto out;
	}
	log(LOG_DEBUG, "user has uid: %d, gid: %d", uid, gid);

	get_cc_name(cache_file5, "/tmp/krb5cc_%d_XXXXXX", uid);
	snprintf(krb5cc, sizeof(krb5cc), "FILE:%s", cache_file5);
	log(LOG_DEBUG, "cache file name: %s", cache_file5);
	if ( (krc = krb5_cc_resolve(context, krb5cc, &ccache)) != KRB5_SUCCESS )
	{
		log(LOG_CRIT, "error resolving credentials cache file: %s",
			error_message(krc));
		goto out;
	}
	log(LOG_DEBUG, "krb5cc: %s", krb5cc);

	/* fetching init credentials' options */
	krb5_get_init_creds_opt_init(&opt);
	krb5_get_init_creds_opt_set_default_flags(context, "pam_krb5keytab", 
		/* XXX */principal->realm, &opt);
	krb5_get_init_creds_opt_set_anonymous(&opt, 0);

	log(LOG_DEBUG, "realm to get ticket for: %s", principal->realm);
	log(LOG_DEBUG, "keytab file: %s", ktfile);

	/* resolve ticket file where to take the user's key from */
	krc = ktfile ?
		krb5_kt_resolve(context, ktfile, &kt) :
		krb5_kt_default(context, &kt);
	if (krc != KRB5_SUCCESS)
	{
		log(LOG_CRIT, "error resolving keytab file: %s", error_message(krc));
		prc = PAM_SYSTEM_ERR;
		goto out;
	}
	cred.client = principal;
	/* now fetch credentials */
	krc = krb5_get_init_creds_keytab(context,
		&cred,
		principal,
		kt,
		start_time,
		NULL,
		&opt);
	if ( krc != KRB5_SUCCESS )
	{
		log(LOG_CRIT, "failed getting initial credentials from keytab (%s)",
			error_message(krc));
		prc = PAM_SYSTEM_ERR;
		goto out;
	}
	krb5_kt_close(context, kt);

	/* initialize cc */
	if ( (krc = krb5_cc_initialize(context, ccache, cred.client))
		!= KRB5_SUCCESS )
	{
		log(LOG_CRIT, "failed initializing credentials (%s)",
			error_message(krc));
		prc = PAM_SYSTEM_ERR;
		goto out;
	}

	/* store cc to file */
	if ( (krc = krb5_cc_store_cred(context, ccache, &cred)) != KRB5_SUCCESS )
	{
		log(LOG_CRIT, "failed storing credentials (%s)", error_message(krc));
		prc = PAM_SYSTEM_ERR;
		goto out;
	}

#ifdef KRB4
	if ( krb5_config_get_bool_default(context, NULL, FALSE, "libdefaults",
		"krb4_get_tickets", NULL) )
	{
		log(LOG_DEBUG, "will try to convert to v4 credentials");
		/* Convert v5creds to v4creds */
		krc = krb524_convert_creds_kdc_ccache(context, ccache, &cred, &v4cred);
		if ( krc != KRB5_SUCCESS )
		{
			log(LOG_CRIT, "error converting creds: %s", error_message(krc));
			goto out;
		}

		/* Build v4cc name */
		get_cc_name(cache_file4, "/tmp/tkt_%d_XXXXXXX", uid);
		log(LOG_DEBUG, "v4 credentials file: %s", cache_file4);
		krb_set_tkt_string(cache_file4);

		/* Store credentials in the ticket file. */
		krc = tf_setup(&v4cred, v4cred.pname, v4cred.pinst);
		if ( krc != KRB5_SUCCESS )
		{
			log(LOG_CRIT, "error saving v4creds: %s", krb_get_err_text(krc));
			goto out;
		}

		chown(cache_file4, uid, gid);
		/* Set up the environment variable for Kerberos 4. */
		snprintf(krb4cc, MAXLEN,
			 "KRBTKFILE=%s", cache_file4);
		pam_putenv(pamh, krb4cc);
	}
#endif
#ifdef AFS
	if ( k_hasafs() )
	{
		k_setpag();
		log(LOG_DEBUG, "generating AFS token");
		krb5_afslog(context, ccache, NULL, NULL);
	}
#endif
	krb5_cc_close(context, ccache);
	chown(cache_file5, uid, gid);

	snprintf(krb5cc, MAXLEN, "KRB5CCNAME=%s",
		cache_file5);
	pam_putenv(pamh, krb5cc);

out:
#ifdef KRB4
	memset(&v4cred, 0, sizeof(v4cred));
#endif
	krb5_free_creds_contents(context, &cred);
	if (principal != NULL) krb5_free_principal(context, principal);
	if (context != NULL) krb5_free_context(context);
	return prc;
}

int delete_creds(pam_handle_t *pamh)
{
	int krc = KRB5_SUCCESS, prc = PAM_SUCCESS;
	char cache_name[] = "KRB5CCNAME", *file_cache = NULL;
	krb5_context context;
	krb5_principal principal;
	krb5_ccache ccache;
	krb5_creds cred;

	log(LOG_DEBUG, "called delete_creds");
	if ( (krc = krb5_init_context(&context)) != KRB5_SUCCESS )
		return PAM_SYSTEM_ERR;

	file_cache = pam_getenv(pamh, cache_name);
	krc = krb5_cc_resolve(context, file_cache, &ccache);
	if (krc != KRB5_SUCCESS)
	{
		log(LOG_CRIT, "error resolving credentials cache file %s (%s)",
			cache_name, error_message(krc));
		prc = PAM_SYSTEM_ERR;
		goto out;
	}
	log(LOG_DEBUG, "deleting cache file %s", file_cache);

	if ( (krc = krb5_cc_destroy(context, ccache)) != KRB5_SUCCESS )
	{
		log(LOG_CRIT, "error destroying credentials cache: %s",
			error_message(krc));
		prc = PAM_SYSTEM_ERR;
		goto out;
	}
#ifdef AFS
	if ( k_hasafs() ) k_unlog();
#endif
#ifdef KRB4
	dest_tkt();
#endif
	
out:
	krb5_free_creds_contents(context, &cred);
	if (context != NULL) krb5_free_context(context);
	return prc;
}

int get_cc_name(char *cc, const char *pattern, ...)
{
	int tmpfd;
	va_list plist;

	memset(cc, 0, MAXLEN);

	va_start(plist, pattern);
	vsnprintf(cc, MAXLEN, pattern, plist);
	va_end(plist);

	tmpfd = mkstemp(cc);
	close(tmpfd);
	tmpfd = -1;

	return 0;
}

/* Function for determining a user's UID and primary GID.  glibc and most
 * other systems version. */
int get_pw(const char *user, uid_t *uid, gid_t *gid)
{
	static struct passwd rec;
	struct passwd *pwd = NULL;
	static char buf[MAXLEN];
	memset(&rec, 0, sizeof(rec));
	if (getpwnam_r(user, &rec, buf, sizeof(buf), &pwd) == 0)
	{
		if (pwd == &rec) {
			*uid = pwd->pw_uid;
			*gid = pwd->pw_gid;
			return TRUE;
		}
	}
	return FALSE;
}

void log(int priority, const char *msg, ...)
{
	char buf[MAXLEN];
	va_list plist;
	va_start(plist, msg);
	vsnprintf(buf, sizeof(buf), msg, plist);
	va_end(plist);
	openlog(SERVICENAME, LOG_PID | LOG_CONS | LOG_NOWAIT, LOG_AUTH);
	syslog(priority, buf);
	closelog();
}

